from .supervisor import Supervisor as supervise
from typing import Union, Callable, Any
from .protocols import PCancelToken


def run(*args: Union[Callable[[], Any], Callable[[PCancelToken], Any]]):
    return supervise(*args).run()



from __future__ import annotations
from asy.protocols import PCancelToken

import asyncio
import logging
import signal
from typing import Any, List, Set, Tuple, Callable, Union
import logging
from .tokens import CancelToken
from .normalizer import normalize_to_schedulable

logger = logging.getLogger(__name__)


class SupervisorBase:
    def __init__(
        self, *schedulables: Union[Callable[[], Any], Callable[[PCancelToken], Any]]
    ):
        tmp = [normalize_to_schedulable(x) for x in schedulables]
        self.schedulables = tmp
        self.set_config()
        self.__post_init__()

    def __post_init__(self):
        pass

    def set_config(
        self, on_succeed=None, on_error=None, on_cancel=None, on_completed=None
    ):
        self.on_succeed = on_succeed or (
            lambda task: logger.info(f"[SUCCESS]{task.result()}")
        )
        self.on_error = on_error or (lambda task: logger.info(f"[FAIL]{task}"))
        self.on_cancel = on_cancel or (lambda task: logger.info(f"[CANCEL]{task}"))
        self.on_completed = on_completed or (
            lambda task: logger.info(f"[COMPLETE]{task}")
        )

    def set_config_print_log(self):
        self.on_succeed = lambda task: print(task)
        self.on_error = lambda task: print(task)
        self.on_cancel = lambda task: print(task)
        self.on_completed = lambda task: print(task)

    def schedule(self) -> Tuple[PCancelToken, asyncio.Task]:
        token = CancelToken()
        task = asyncio.create_task(self.__call__(token))
        return token, task

    def run(self, handle_signals: Set[str] = {"SIGINT", "SIGTERM"}):
        """新たなイベントループ上に、管理している関数群をスケジューリングし、完了まで監督する。このメソッドは自身の状態を変更しない。"""
        loop = asyncio.new_event_loop()
        token = CancelToken()

        def handle_cancel():
            token.is_cancelled = True

        for sig_name in handle_signals:
            sig = getattr(signal, sig_name)
            loop.add_signal_handler(sig, handle_cancel)

        try:
            result = loop.run_until_complete(self(token))
        except Exception as e:
            raise
        finally:
            loop.close()

        return result

    async def __call__(self, token: PCancelToken):
        """管理している関数群をスケジューリングし、完了まで監督する。このメソッドは自身の状態を変更しない。"""
        tokens, tasks, future, sub_futures = self._start()

        async def observe_cancel():
            while not future.done():
                await asyncio.sleep(0.1)
                if token.is_cancelled:
                    for token in tokens:
                        token.is_cancelled = True

        task = asyncio.create_task(observe_cancel())
        sub_futures.append(task)
        result = await future
        return await self.finalize(result, sub_futures)

    @classmethod
    async def finalize(cls, wait_result, sub_futures):
        for sub in sub_futures:
            sub.cancel()

        result = cls.get_result_from_wait(wait_result)
        await asyncio.sleep(0)
        return result

    @staticmethod
    def get_result_from_wait(wait_result):
        return wait_result[0]

    def _start(self):
        schedulables = self.schedulables
        on_succeed = self.on_succeed
        on_error = self.on_error
        on_cancel = self.on_cancel
        on_completed = self.on_completed

        def on_done(task):
            try:
                result = task.result()
                on_succeed(task)
            except asyncio.CancelledError:
                # logger.info(f"[CANCEL]{task}")
                on_cancel(task)
            except Exception:  # pylint: disable=broad-except
                # logger.exception(message, *message_args)
                # import traceback

                # logger.warning(traceback.format_exc())
                # logger.warning(f"[FAILED]{task}")
                on_error(task)
            on_completed(task)

        tasks = []
        tokens = []
        sub_futures = []

        for index, schedulable in enumerate(schedulables):
            token, task = schedulable.schedule()
            tokens.append(token)
            tasks.append(task)
            task.add_done_callback(on_done)

        future = asyncio.create_task(
            asyncio.wait(tasks, return_when=asyncio.ALL_COMPLETED)
        )
        return tokens, tasks, future, sub_futures


class Supervisor(SupervisorBase):
    def __post_init__(self):
        self.future: asyncio.Future = None  # type: ignore
        self.cancel_tokens: List[PCancelToken] = None  # type: ignore
        self.tasks = None
        self.sub_futures = None

    async def start(self):
        if not self.is_ready:
            raise Exception("already running.")

        tokens, tasks, future, sub_futures = self._start()
        self.future = future
        self.cancel_tokens = tokens
        self.tasks = tasks
        self.sub_futures = sub_futures
        return tokens, tasks, future, sub_futures

    def cancel(self):
        if self.is_ready:
            raise Exception("The coroutine has not been executed yet")

        for token in self.cancel_tokens:
            token.is_cancelled = True

    async def stop(self, timeout=10000):
        self.cancel()
        await self.future

    def is_completed(self):
        return self.future.done()

    @property
    def is_ready(self):
        return all(
            [
                self.future is None,
                self.cancel_tokens is None,
                self.tasks is None,
                self.sub_futures is None,
            ]
        )

    def clear(self):
        self.future = None
        self.cancel_tokens = None
        self.tasks = None
        self.sub_futures = None
        assert self.is_ready



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['asy']

package_data = \
{'': ['*']}

install_requires = \
['typer>=0.3.2,<0.4.0']

setup_kwargs = {
    'name': 'asy',
    'version': '0.0.5',
    'description': 'asy is easy and powerful supervisor for asyncio.',
    'long_description': '# asy\n[![Version](https://img.shields.io/pypi/v/asy)](https://pypi.org/project/asy)\n[![License: MIT](https://img.shields.io/badge/license-MIT-yellow.svg)](https://opensource.org/licenses/MIT)\n\n`asy` is easy and powerful supervisor for `asyncio`.\n\n# Motivation for development\n\n- Simple cancellation\n- Improve the coordination of async functions between libraries\n- No more programs for execution management\n- Develop specifications like ASGI\n\n# Requirement\n\n- Python 3.8+\n\n# Installation\n\n``` shell\npip install asy\n```\n\n# Getting started\n\nCreate functions in `example.py`:\n\nAll you have to do is say the magic word `token`, and you can handle the function\'s lifetime at will.\n\n``` python\nimport asyncio\n\n# cancelable infinity loop\nasync def func1(token):\n    while not token.is_cancelled:\n        await asyncio.sleep(1)\n    return "complete func1."\n\n\n# uncancelable limited loop\nasync def func2(token):\n    for i in range(10):\n        await asyncio.sleep(1)\n    return f"complete func2.  result: {i}"\n\n\n# force cancel infinity loop\nasync def func3():\n    while True:\n        await asyncio.sleep(1)\n    return "complete func3. unreachable code."\n\n\n# uncancelable limited loop\ndef func4():\n    for i in range(1000):\n        ...\n    return f"complete func4.  result: {i}"\n\n# from callable\nclass YourDeamon:\n    def __init__(self, value):\n        self.value = value\n\n    async def __call__(self, token):\n        value = self.value\n\n        while not token.is_cancelled:\n            await asyncio.sleep(1)\n        return f"complete func5.  result: {value}"\n\nfunc5 = YourDeamon()\n\n# Do not run\n# infinity loop\n# async def func5()):\n#     while True:\n#         print("waiting")\n```\n\nRun in shell.\n\n``` shell\npython3 -m asy example:func1 example:func2 example:func3 example:func4 example:func5\n```\n\nRun in Python script.\n\n``` python\nimport asy\nfrom example import func1, func2, func3, func4, func5\n\nsupervisor = asy.supervise(func1, func2, func3, func4, func5)\nsupervisor.run()\n\n# or\nasy.run(func1, func2, func3, func4, func5)\n```\n\n\nLet\'s end the daemon with `Ctrl-C` and enjoy `asy`!\n\n# Caution\n`asy` is a beta version. Please do not use it in production.\n',
    'author': 'sasano8',
    'author_email': 'y-sasahara@ys-method.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/sasano8/asy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
